/*
  Name: PcmFlash.c
  Copyright: 
  Author: Darren Freed
  Date: 03/08/04 21:11
  Description: 
      This file can be used to reflash the '7058 or '1539 pcms via the SCI (not DLC2) using mode 10
      the user has the option to specify which block to reprogram
      Program execution: the user specifies the options and input files
            1) the program then downloads the reprogramming algorithm to pcm ram
            2) the downloaded reprogramming algorithm erases the flash
            3) the program then writes the new flash data to the pcm - 
            4) msg structure: $F4,msg length,$06(MODE),subroutine start address, data start address,#of data bytes, data, cksum
            5) after each msg is sent and programmed, the ecm returns the flash status
*/


#include <stdio.h>              /* Standard input/output definitions */
#include <string.h>             /* String function definitions */
#include <unistd.h>             /* UNIX standard function definitions */
#include <fcntl.h>              /* File control definitions */
#include <errno.h>              /* Error number definitions */
#include <termios.h>            /* POSIX terminal control definitions */
#include <sys/select.h>
#include "pcmflash.h"

#define BUF_SIZE 0x100                      /* Size of Rx and Tx buffers  */
#define SIZE 0xac                           /* for defining the size of each message */
#define FALSE 0
#define TRUE  !FALSE
#define BAUD  B19200

int read_file (char *buffer,unsigned size);
void restore  (void);                           /* restore COM vector   */
void put_com  (unsigned char c);                /* transmit data        */
int erase_pcm ();
int program_pcm (unsigned startaddr, unsigned size);
void get_flash_id(void);
void load_pcm(void);
void reset_pcm(void);

/* The following are treated as global variables and are  */
/* declared outside the main program.                     */

unsigned char transmitting = FALSE;
unsigned char receiving = FALSE;
unsigned char rx_buff[BUF_SIZE];           /* receive data buffer  */
unsigned char tx_buff[BUF_SIZE];           /* transmit data buffer  */
int rx_in_ptr, rx_ptr;                   /* rx pointer*/
int tx_ptr;                              /* tx pointer*/
int size;                                   /* the transmit msg length */
int com_base;                              /* COM port base address */
unsigned char cksum;                        /* transmit message cksum */
unsigned startaddr, count;
unsigned blockaddr;                        //this is the number the user enters
unsigned char *offset;                     //this is the source file 
unsigned short flash_mfg, flash_dev;
unsigned flashfile_size,flashsize;
unsigned blocksize;                        //this is the size of the block to be reflashed
char *flash_buffer;


int main (int argc, char *argv[])
{  
  char *flashname;
  FILE *flashfile;
  struct stat filestat;


  if (argc<2){
    printf("\nUsage:  pcmflash filename block\n\n");
    printf("filename is a binary image to be loaded into the flash\n");
    printf("block is the flash block byte address to be programmed:\n");
    printf(" 0 -  16kb boot block \n");
    printf(" 4 -   8kb parameter block 1 \n");
    printf(" 6 -   8kb parameter block 2 \n");
    printf(" 8 -  96kb main block \n");
    printf("20 - 128kb main block \n");
    printf("40 - 128kb main block (512k chip) \n");
    printf("60 - 128kb main block (512k chip) \n");
    printf("If no block is specified, the entire chip will be programmed.\n");
    printf("The image size will be compared to the flash size to make sure it matches.\n");
    exit(1);
    }    
  flashname=argv[1];
  if (argc<3) {
      blockaddr=1;                         // this will cause the default to be used in the switch statement
  }    

  else {
      sscanf(argv[2],"%x", &blockaddr);    // blockaddr is a hex number
  }

  flashfile=fopen (flashname,"rb");
  if (!flashfile){
      printf("Unable to open file %s.\n",argv[1]);    
      exit(1);
  }    
  
  stat(flashname,&filestat);
  flashfile_size=filestat.st_size;

  if (init_com( BAUD )){             /* initialize COM port */
      load_pcm();
      get_flash_id();
      flashsize=get_flash_size(flash_mfg,flash_dev);
      if (flashfile_size==flashsize) {
          flashsize=flashfile_size;  // for troubleshooting only
          printf("Reading flash file '%s' (0x%x bytes)\n",flashname,flashfile_size);
          flash_buffer=malloc(flashfile_size);
          if (flash_buffer) {
              fread(flash_buffer,flashfile_size,1,flashfile);
              erase_pcm(blockaddr,flashsize);
              program_pcm(blockaddr,blocksize); 
              reset_pcm();
          }
          else  {
              printf("Unable to allocate %d bytes for a flash buffer.\n",flashfile_size);
          }    
          fclose(flashfile);
          restore();                      // restore com settings
          printf("All Done !!\n");              
      }
      else {
          printf("Flash file size (0x%x) does not match flash size (0x%x).\n",flashfile_size,flashsize);
      }
  }    
  else{
       printf("Unable to open com port.\n");        
   }    
  exit(1);
}


/* THIS IS FOR SENDING THE 'LOADER' ALGORITHM TO THE PCM */
void load_pcm(void)
{
  char *algoname;
  FILE *algofile;
  char algo_buffer[0x408];                 /* buffer to contain the algo */
  unsigned x, algofile_size, start;
  char *offset;
  int count;

  if ((algofile=fopen("flash.bin","rb")) != NULL){
     fread(algo_buffer,sizeof(unsigned char),0x408,algofile);
     printf("Initializing PCM...\n");
     for (count = 0; count < 6; count++) {
         tx_ptr = 0;                                 /* reset tx pointer */    
         cksum = 0;
         size = SIZE;                                /* total msg length,  */
         start=(size*count);                                 
         offset = &algo_buffer[start];                                 /* point to the correct place in the algo file buffer */
         read_file(offset,size);                                       /* read the algo file and send it to tx_buff */
         cksum=~cksum+1;
         tx_buff[tx_ptr] = cksum;
         tx_ptr+=1;
         transmit(tx_buff,tx_ptr);
       /********************************
       /*  NOW SWITCH TO RECEIVE MODE  *
       /*******************************/
//         printf("\nWaiting for response from PCM\n");
         rx_ptr = 0;
         receiving = TRUE;
         x = 0;
         do {
             rx_ptr += receive(rx_buff,BUF_SIZE);            // get number of characters received, add so that we don't miss any
             if (rx_ptr > tx_ptr) {                               // if the received message was more than just the echo
//                 printf("\nReceived Msg: $%02x,$%02x,$%02x,$%02x,$%02x\n",rx_buff[tx_ptr+1],rx_buff[tx_ptr+2],rx_buff[tx_ptr+3],rx_buff[tx_ptr+4],rx_buff[tx_ptr+5]);   /* print it */
                 receiving = FALSE;
                 rx_ptr = 0;                                              /* reset receive pointer */
                 break;
                 }    
             else if (rx_ptr <= tx_ptr){                           // if the received message was just the echo
                 if (x == 5){
                     printf("\n\nNo Response from PCM\n");
                     restore();
                     exit(1);
                     }
                 printf("\nstill waiting...\n");
                 x++;
                 usleep(100000);
                }   
             }    
         while (receiving == TRUE);
         }    
       }    
   else {printf("Unable to open algo file.\n");}
   return;
}

// for reading the loader.bin file and the new flash image
int read_file(char *buffer,unsigned size) 
{
    unsigned addr;
    unsigned char b, running_cksum=0;
    unsigned char counter;                      /* counter for the running msg length */
    counter = 0;
    while(counter<size){
        b=*buffer;
        put_com(b);                   /* write to transmit buffer */
        counter++;
        buffer++;
    }
    return;
}

void put_com(unsigned char c)                      /* Fill Tx Buffer       */
{
      tx_buff[tx_ptr] = c;                /* write data to transmit buffer  */
      tx_ptr = tx_ptr+1;                  /* and update tx buffer input ptr */
      cksum+=c;                           /* and add to running cksum */      
}

void get_flash_id(void)
{
    int i=0;
    tx_ptr = 0;
    cksum=0;
    printf("Getting flash ID...\n");
    put_com(0xf4);
    put_com(0x61);
    put_com(0x10);
    put_com(0xff);
    put_com(0x80);
    put_com(0x00);
    put_com(0x4e);
    put_com(0xb9);
    put_com(0x00);
    put_com(0xff);
    put_com(0x83);
    put_com(0x3e);
    put_com(0x4e);
    put_com(0x75);
    cksum=~cksum+1;
    tx_buff[tx_ptr] = cksum;
    transmit(tx_buff,tx_ptr+1);
    receiving = TRUE;
    rx_ptr += receive(rx_buff,BUF_SIZE);            // get number of characters received, add so that we don't miss any
    do{
        if(rx_ptr > tx_ptr){
            flash_mfg = ((rx_buff[tx_ptr+3])<<8)+(rx_buff[tx_ptr+4]);
            flash_dev = ((rx_buff[tx_ptr+5])<<8)+(rx_buff[tx_ptr+6]);
            printf("Flash manufacturer ID=0x%x, device code=0x%x\n",flash_mfg,flash_dev);
            receiving = FALSE;
            rx_ptr = 0;
        }
        else i ++;
        usleep (10000);
        if (i > 10){
            printf("Unable to identify flash part - no response from pcm\n");
            receiving == FALSE;                        // stop trying
            exit(1);
        }    
    }
    while (receiving == TRUE);    
  return;    
}    

/* THIS IS FOR ERASING THE PCM */
int erase_pcm()
{
  switch(blockaddr){
      case 0x0 :
        blocksize=0x4000;                            // byte size (not word size)
        printf("Erasing 16kb boot block\n");
          erase_block(0x0);               // 16kb boot block
          break;
      
      case 0x4000 :
        blocksize=0x2000;                            // byte size (not word size)
        printf("Erasing 8k parameter block\n");
          erase_block(0x4000);            // 8k parameter block
          break;  
      
      case 0x6000 :
        blocksize=0x2000;                            // byte size (not word size)
        printf("Erasing 8k parameter block\n");
          erase_block(0x6000);            // 8k parameter block
          break;

      case 0x8000 :
        blocksize=0x18000;                            // byte size (not word size)
        printf("Erasing 96k main block\n");
          erase_block(0x8000);            // 96k main block
          break;

      case 0x20000 :
        blocksize=0x20000;                            // byte size (not word size)
        printf("Erasing 128k main block\n");
          erase_block(0x20000);           // 128k main block
          break;
                                    
      case 0x40000 :
          if (flashsize == 0x80000){
              blocksize=0x20000;                            // byte size (not word size)
              printf("Erasing 128k main block\n");
              erase_block(0x40000);           // 128k main block
          }
          else printf("Invalid address %x for flash part\n",blockaddr);
          break;

      case 0x60000 :
          if (flashsize == 0x80000){
              blocksize=0x20000;                            // byte size (not word size)
              printf("Erasing 128k main block\n");
              erase_block(0x60000);           // 128k main block
          }    
          else printf("Invalid address %x for flash part\n",blockaddr);
          break;

      default :
          switch(flashsize){
              case 0x40000 :
                      blockaddr=0x0;                         // reset to zero
                      blocksize=flashsize;
                      printf("Erasing 16kb boot block\n");
                      erase_block(0x0);               // 16kb boot block
                      printf("Erasing 8k parameter block\n");
                      erase_block(0x4000);            // 8k parameter block
                      printf("Erasing 8k parameter block\n");
                      erase_block(0x6000);            // 8k parameter block
                      printf("Erasing 96k main block\n");
                      erase_block(0x8000);            // 96k main block
                      printf("Erasing 128k main block\n");
                      erase_block(0x20000);           // 128k main block
                      break;
                      
               case 0x80000 :
                      blockaddr=0x0;                         // reset to zero
                      blocksize=flashsize;
                      printf("Erasing 16kb boot block\n");
                      erase_block(0x0);               // 16kb boot block
                      printf("Erasing 8k parameter block\n");
                      erase_block(0x4000);            // 8k parameter block
                      printf("Erasing 8k parameter block\n");
                      erase_block(0x6000);            // 8k parameter block
                      printf("Erasing 96k main block\n");
                      erase_block(0x8000);            // 96k main block
                      printf("Erasing 128k main block\n");
                      erase_block(0x20000);           // 128k main block
                      printf("Erasing 128k main block\n");
                      erase_block(0x40000);           // 128k main block
                      printf("Erasing 128k main block\n");
                      erase_block(0x60000);           // 128k main block
                      break;
                  }    
              }    
  return;
}  


int erase_block(unsigned addr)
{
    unsigned char addrlo, addrmid;
    unsigned char addrhi;
    char y,n;
    printf("Erasing PCM...\n");
    cksum = 0;
    tx_ptr = 0;
    addrlo = (addr & 0xff);                 
    addrmid= (addr >> 8);
    addrhi = (addr >> 16);
    put_com(0xF4);
    put_com(0x65);                           /* msg length */
    put_com(0x10);
    put_com(0xff);
    put_com(0x80);
    put_com(0x00);
    put_com(0x4e);
    put_com(0xb9);
    put_com(0x00);                        // address of erase routine
    put_com(0xff);
    put_com(0x83);
    put_com(0x62);
    put_com(0x4e);
    put_com(0x75);
    put_com(0x00);                           /* address to be erased */
    put_com(addrhi);
    put_com(addrmid);
    put_com(addrlo);
    cksum=~cksum+1;
    tx_buff[tx_ptr] = cksum;
    tx_ptr+=1;
    transmit(tx_buff,tx_ptr);
    rx_ptr = 0;
    y = 0;
    receiving = TRUE;
    do {
        rx_ptr += receive(rx_buff,BUF_SIZE);            // get number of characters received, add so that we don't miss any
        if (rx_ptr > tx_ptr){
//            printf("\nReceived Msg: $%02x,$%02x,$%02x,$%02x,$%02x\n",rx_buff[tx_ptr],rx_buff[tx_ptr+1],rx_buff[tx_ptr+2],rx_buff[tx_ptr+3],rx_buff[tx_ptr+4]);   /* print it */
            receiving = FALSE;
            rx_ptr = 0;                                              /* reset receive pointer */
            }    
        else if (rx_ptr <= tx_ptr){
            if (y == 20){
                printf("\n\nNo Response from PCM during erase.\n");
                restore();
                exit(1);
            }
            y++;
            }   
        }    
      while (receiving == TRUE);
      return;
}        


int program_pcm(unsigned startaddr,unsigned size)
{
    unsigned addr;
    unsigned char addrlo, addrmid, addrhi;
    char y,n;
    printf("Programming PCM...\n");
    for (addr=startaddr;addr<startaddr+size;addr+=0x80) {
      if ((addr & 0xfff)==0){
          printf("Programmed address 0x%08x\n",addr);
      }    
    cksum = 0;
    tx_ptr = 0;
    addrlo = (addr & 0xff);                 
    addrmid= (addr >> 8);
    addrhi = (addr >> 16);
    put_com(0xF4);
    put_com(0xe7);                           /* msg length */
    put_com(0x10);
    put_com(0xff);
    put_com(0x80);
    put_com(0x00);
    put_com(0x4e);
    put_com(0xb9);
    put_com(0x00);
    put_com(0xff);
    put_com(0x83);
    put_com(0x92);
    put_com(0x4e);
    put_com(0x75);
    put_com(0x00);                           /* address */
    put_com(addrhi);
    put_com(addrmid);
    put_com(addrlo);
    put_com(0x00);
    put_com(0x80);                        // number of bytes to program
    offset=&flash_buffer[addr];               // offset points to the source flash file    
    read_file(offset,0x80);
    cksum=~cksum+1;
    tx_buff[tx_ptr] = cksum;
    tx_ptr+=1;
    transmit(tx_buff,tx_ptr);
    rx_ptr = 0;
    y = 0;
    receiving = TRUE;
    usleep(1000);
    do {
        rx_ptr += receive(rx_buff,BUF_SIZE);            // get number of characters received, add so that we don't miss any
        if (rx_ptr > tx_ptr){
//            printf("\nReceived Msg: $%02x,$%02x,$%02x,$%02x,$%02x\n",rx_buff[tx_ptr],rx_buff[tx_ptr+1],rx_buff[tx_ptr+2],rx_buff[tx_ptr+3],rx_buff[tx_ptr+4]);   /* print it */
            receiving = FALSE;
            rx_ptr = 0;                                              /* reset receive pointer */
            }    
        else if (rx_in_ptr <= tx_ptr){
            if (y == 20){
                printf("\n\nNo Response from PCM during programming.\n");
                restore();
                exit(1);
            }
            y++;
            }   
        }    
      while (receiving == TRUE);
  }                                                      // end of for statement
  return;
}        

void reset_pcm(void)
{
    printf("Resetting PCM...\n");
    cksum = 0;
    tx_ptr = 0;
    put_com(0xf4);
    put_com(0x61);
    put_com(0x10);
    put_com(0xff);
    put_com(0x80);
    put_com(0x00);
    put_com(0x4e);
    put_com(0xb9);
    put_com(0x00);
    put_com(0xff);
    put_com(0x83);
    put_com(0xd0);
    put_com(0x4e);
    put_com(0x75);
    cksum=~cksum+1;
    tx_buff[tx_ptr] = cksum;
    transmit(tx_buff,tx_ptr+1);
}    

